import wx
import os, shutil
from freemet.config import config

_ = wx.GetTranslation

class SetDialog(wx.Dialog):
    BTN_REMOVE_ALL  = wx.NewId()
    BTN_REMOVE      = wx.NewId()
    BTN_ADD         = wx.NewId()
    BTN_ADD_ALL     = wx.NewId()
    BTN_NEW_SET     = wx.ID_NEW
    BTN_DELETE_SET  = wx.ID_DELETE
    BTN_SAVE_SET    = wx.ID_SAVE
    BTN_RENAME_SET  = wx.ID_EDIT

    def __init__(self, parent, sets, currentSet, available):
        wx.Dialog.__init__(self, parent, -1, title=_("Set Dialog"))

        # Capture closing event
        self.Bind(wx.EVT_CLOSE, self.OnClose)

        # All available modules
        self.available = dict(available)

        # _Copy_ set dictionary to still having the original one when hitting
        # the cancel button
        self.sets = dict(sets)

        self.savedSets = dict(sets)
        self.modified = False
        
        self.reverseNameMap = {}
        for key, name in self.available.iteritems():
            self.reverseNameMap[name] = key

        # Sizers
        mainsizer           = wx.BoxSizer(wx.VERTICAL)
        listssizer          = wx.FlexGridSizer(2, 3)
        listbuttonsizer     = wx.BoxSizer(wx.VERTICAL)
        setmgrsizer         = wx.BoxSizer(wx.HORIZONTAL)
        setmgrbuttonsizer   = wx.FlexGridSizer(2, 2)
        okcancelsizer       = self.CreateButtonSizer(wx.OK | wx.CANCEL)

        btnOk = self.FindWindowById(wx.ID_OK)
        assert btnOk
        btnCancel = self.FindWindowById(wx.ID_CANCEL)
        assert btnCancel
        self.Bind(wx.EVT_BUTTON, self.OnButtonOk, btnOk)
        self.Bind(wx.EVT_BUTTON, self.OnButtonCancel, btnCancel)

        # Controls
        # Build combobox this early, because most operations rely on
        # self.comboboxSets.GetSelection()
        self.comboboxSets = wx.ComboBox(self, -1, style=wx.CB_READONLY)
        self.__RebuildCombobox(self.comboboxSets, self.sets)
        self.comboboxSets.SetValue(currentSet)
        self.oldSetname = self.comboboxSets.GetValue()

        # Set lists (available and active)
        listssizer.Add(wx.StaticText(self, -1, _("Available:")))
        listssizer.AddSpacer(10)
        listssizer.Add(wx.StaticText(self, -1, _("Active:")))
        self.listboxAvailable = wx.ListBox(self, -1, size=(250, 200),
                                           choices=self.__MakeAvailableList())
        self.listboxActive = wx.ListBox(self, size=(250, 200),
                                        choices=self.__MakeActiveList())
        listssizer.Add(self.listboxAvailable, 1, 1)
        listssizer.Add(listbuttonsizer)
        listssizer.Add(self.listboxActive)

        # Buttons to control set lists
        for x in [("<<", self.BTN_REMOVE_ALL),
                  ("<-", self.BTN_REMOVE),
                  ("->", self.BTN_ADD),
                  (">>", self.BTN_ADD_ALL)]:
            btn = wx.Button(self, x[1], x[0], size=(30, 30))
            listbuttonsizer.Add(btn, 0, wx.ALL, 10)
            self.Bind(wx.EVT_BUTTON, self.OnListButton, btn)

        # Set management
        setmgrsizer.Add(wx.StaticText(self, -1, _("Set:")),
                        0, wx.ALIGN_CENTER_VERTICAL)
        self.Bind(wx.EVT_COMBOBOX, self.OnComboboxSets)
        setmgrsizer.Add(self.comboboxSets, 0, wx.ALIGN_CENTER_VERTICAL)
        setmgrsizer.Add(setmgrbuttonsizer, 0, wx.ALL, 10)

        # Buttons to manage sets
        for x in [(_("New set"), self.BTN_NEW_SET),
                  (_("Delete set"), self.BTN_DELETE_SET),
                  (_("Save set"), self.BTN_SAVE_SET),
                  (_("Rename set"), self.BTN_RENAME_SET)]:
            btn = wx.Button(self, x[1])#, x[0])
            setmgrbuttonsizer.Add(btn)
            self.Bind(wx.EVT_BUTTON, self.OnSetMgrButton, btn)
        
        # Disable save button (no changes yet)
        self.__DisableButtons([self.BTN_SAVE_SET])

        # Put everything together
        mainsizer.Add(listssizer, 0, wx.ALL, 10)
        mainsizer.Add(setmgrsizer, 0, wx.ALL, 10)
        mainsizer.Add(okcancelsizer, 0, wx.ALL | wx.ALIGN_RIGHT, 10)
        self.SetSizer(mainsizer)

        self.Fit()
        
    def GetSetList(self):
        return dict(self.savedSets)
    
    def GetSetName(self):
        return self.comboboxSets.GetValue()

    def OnClose(self, evt):
        if self.modified:
            r = self.__ShowUnsavedDataQuestion()
            if r == wx.YES:
                self.savedSets = dict(self.sets)
                self.modified = False
                self.Destroy()
            elif r == wx.NO:
                self.modified = False
                self.Destroy()
            elif r == wx.CANCEL:
                pass

        self.EndModal(wx.ID_CANCEL)

    def OnButtonOk(self, evt):
        self.__MarkCurrentSetUnmodified()
        self.savedSets = dict(self.sets)
        self.EndModal(wx.ID_OK)

    def OnButtonCancel(self, evt):
        self.EndModal(wx.ID_CANCEL)

    def OnListButton(self, evt):
        id = evt.GetId()
        
        if id == self.BTN_REMOVE_ALL:
            # Move active plugins to available list
            available = self.__MakeAvailableList() + self.__MakeActiveList()
            available.sort()
            active = []
            
            setname = self.comboboxSets.GetValue()
            self.sets[setname] = []

            # Clear and repopulate available listbox
            self.__RebuildListbox(self.listboxAvailable, available)

            # Clear active listbox
            self.listboxActive.Clear()

            # Disable all buttons that move stuff to the active list
            self.__DisableButtons([self.BTN_REMOVE_ALL, self.BTN_REMOVE])

            # Enable all buttons that move stuff to the available list
            self.__EnableButtons([self.BTN_ADD_ALL, self.BTN_ADD])

            self.modified = True
            self.__MarkCurrentSetModified()
        
        elif id == self.BTN_REMOVE:
            sel = self.listboxActive.GetSelection()
            if sel == wx.NOT_FOUND: return

            assert self.listboxActive.GetCount() > 0

            # Get active plugin and remove it
            active = self.__MakeActiveList()
            value = active[sel]
            del active[sel]
            self.listboxActive.Delete(sel)
            
            setname = self.comboboxSets.GetValue()
            self.sets[setname] = []
            for name in active:
                self.sets[setname].append(self.reverseNameMap[name])

            # Put plugin into available list
            available = self.__MakeAvailableList()
            self.__RebuildListbox(self.listboxAvailable, available)

            # If there aren't any entries left in the active list,
            # disable buttons
            if not active:
                self.__DisableButtons([self.BTN_REMOVE_ALL, self.BTN_REMOVE])

            # Enable buttons that move stuff to the available list
            self.__EnableButtons([self.BTN_ADD_ALL, self.BTN_ADD])

            self.modified = True
            self.__MarkCurrentSetModified()

        elif id == self.BTN_ADD:
            sel = self.listboxAvailable.GetSelection()
            if sel == wx.NOT_FOUND: return

            assert self.listboxAvailable.GetCount() > 0

            # Get available plugin and remove it
            available = self.__MakeAvailableList()
            value = available[sel]
            del available[sel]
            self.listboxAvailable.Delete(sel)
            
            # Put plugin into active list
            active = self.__MakeActiveList()
            active.append(value)
            active.sort()
            
            setname = self.comboboxSets.GetValue()
            self.sets[setname] = []
            for name in active:
                self.sets[setname].append(self.reverseNameMap[name])
            
            self.__RebuildListbox(self.listboxActive, active)
            
            # If there aren't any entries left in the available list,
            # disable buttons
            if not available:
                self.__DisableButtons([self.BTN_ADD_ALL, self.BTN_ADD])

            # Enable buttons that move stuff to the active list
            self.__EnableButtons([self.BTN_REMOVE_ALL, self.BTN_REMOVE])
            
            self.modified = True
            self.__MarkCurrentSetModified()
        
        elif id == self.BTN_ADD_ALL:
            # Move available plugins to active list
            active = self.__MakeAvailableList() + self.__MakeActiveList()
            active.sort()
            available = []
            
            setname = self.comboboxSets.GetValue()
            
            #self.sets[setname] = active[:]#[]
            #for x in active: self.sets[setname].append(x)#(self.available)
            self.sets[setname] = []
            for name in active:
                self.sets[setname].append(self.reverseNameMap[name])

            # Clear and repopulate available listbox
            self.__RebuildListbox(self.listboxActive, active)

            # Clear active listbox
            self.listboxAvailable.Clear()

            # Disable all buttons that move stuff to the available list
            self.__DisableButtons([self.BTN_ADD_ALL, self.BTN_ADD])

            # Enable all buttons that move stuff to the active list
            self.__EnableButtons([self.BTN_REMOVE_ALL, self.BTN_REMOVE])

            self.modified = True
            self.__MarkCurrentSetModified()

    def OnSetMgrButton(self, evt):
        id = evt.GetId()

        if id == self.BTN_NEW_SET:
            # Don't create a new set if there is already a brand new one
            try:
                self.sets[_("New*")]
            except KeyError:
                pass
            else:
                msg = _("There is already new set.\nRename it first!")
                wx.MessageBox(msg, _("Error"), wx.OK | wx.ICON_INFORMATION)

            # Create a new empty set
            self.sets[_("New*")] = []
            self.__RebuildCombobox(self.comboboxSets, self.sets)

            self.comboboxSets.SetValue(_("New*"))
            self.oldSetname = _("New*")

            self.__RebuildListbox(self.listboxAvailable, self.available.values())
            self.__RebuildListbox(self.listboxActive, [])
            
            self.__DisableButtons([self.BTN_REMOVE, self.BTN_REMOVE_ALL])
            self.__EnableButtons([self.BTN_ADD, self.BTN_ADD_ALL,
                                  self.BTN_SAVE_SET])
            
            # Create config directory
            try: os.mkdir("config/New")
            except OSError:
                if config["verbose"]:
                    print "Could not create directory config/New"

            self.modified = True
            
        elif id == self.BTN_DELETE_SET:
            setname = self.comboboxSets.GetValue()

            del self.sets[setname]
            
            sel = self.comboboxSets.FindString(setname)
            assert sel != wx.NOT_FOUND
            self.comboboxSets.Delete(sel)

            self.comboboxSets.SetSelection(0)
            self.__RebuildListbox(self.listboxAvailable,
                                  self.__MakeAvailableList())
            self.__RebuildListbox(self.listboxActive,
                                  self.__MakeActiveList())

            #self.savedSets = dict(self.sets)
            self.modified = False
            
            # Delete config directory
            dir = setname
            if setname[-1] == "*": dir = setname[:-1]
            
            try: os.rmdir("config/" + dir)
            except OSError:
                if config["verbose"]:
                    print "Could not delete directory " + "config/" + setname
            
        elif id == self.BTN_SAVE_SET:
            self.modified = False
            self.__MarkCurrentSetUnmodified()
            
            setname = self.comboboxSets.GetValue()
            self.savedSets[setname] = list(self.sets[setname])

        elif id == self.BTN_RENAME_SET:
            setname = self.comboboxSets.GetValue()
            
            dlg = wx.Dialog(self, -1, _("Rename set"))
            sizer = wx.BoxSizer(wx.VERTICAL)
            
            text = wx.TextCtrl(dlg, -1)
            if setname[-1] == "*": text.SetValue(setname[:-1])
            else: text.SetValue(setname)
            
            sizer.Add(text, 1, wx.ALL | wx.EXPAND, 5)
            sizer.Add(dlg.CreateButtonSizer(wx.OK | wx.CANCEL),
                      0, wx.ALL, 5)
            dlg.SetSizer(sizer)
            dlg.Fit()
            
            def OnChar(evt):
                def ResetCtrl():
                    text.SetBackgroundColour(wx.SystemSettings_GetColour(wx.SYS_COLOUR_WINDOW))
                    text.Refresh()
                
                key = evt.GetKeyCode()
                if key >= wx.WXK_SPACE and key != wx.WXK_DELETE and key <= 255:
                    if chr(key) in ["*", "/", "\\"]:
                        text.SetBackgroundColour("pink")
                        text.SetFocus()
                        text.Refresh()
                        wx.CallLater(250, ResetCtrl)
                        return
                evt.Skip()
                
            text.Bind(wx.EVT_CHAR, OnChar)

            r = dlg.ShowModal()

            if r == wx.ID_OK:
                newname = text.GetValue()
                
                # Save set lists before deleting old dictionary entry
                tuple = list(self.sets[setname])
                del self.sets[setname]
                
                # Create new entry
                self.sets[newname] = tuple
                    
                self.modified = True
                self.__MarkCurrentSetModified(newname)

                self.__RebuildCombobox(self.comboboxSets, self.sets)
                self.comboboxSets.SetValue(newname + "*")
                
                # Rename config directory
                olddir = setname
                if setname[-1] == "*":
                    olddir = setname[:-1]
                
                shutil.move("config/" + olddir, "config/" + newname)

            dlg.Destroy()

    def OnComboboxSets(self, evt):
        setname = self.comboboxSets.GetValue()
        
        # No actual new selection (reselection if the already selected one)
        if self.oldSetname == setname: return
        
        if self.modified:
            r = self.__ShowUnsavedDataQuestion()
            if r == wx.YES:
                self.savedSets = dict(self.sets)
                
                self.modified = False
                self.__MarkCurrentSetUnmodified(self.oldSetname, False)
                
            elif r == wx.NO:
                self.modified = False
                
                
                #if self.oldSetname == _("New"):
                try:
                    self.savedSets[self.oldSetname]
                except KeyError:
                    # "New" not found in self.savedSets => delete from sets 
                    del self.sets[self.oldSetname]
                    self.__RebuildCombobox(self.comboboxSets, self.sets)
                else:
                    # Undo changes (restore from self.savedSets)
                    self.sets[self.oldSetname] = list(self.savedSets[self.oldSetname])
                    self.__MarkCurrentSetUnmodified(self.oldSetname, False)

            elif r == wx.CANCEL:
                self.comboboxSets.SetValue(self.oldSetname)

        # A new entry has been selected
        if not self.modified:
            self.oldSetname = setname
            
            # Activate or deactivate list buttons according to available and
            # active lists of the now selected set
            available = self.__MakeAvailableList()
            active = self.__MakeActiveList()
            
            self.__RebuildListbox(self.listboxAvailable, available)
            self.__RebuildListbox(self.listboxActive, active)
            
            addbuttons = [self.BTN_ADD, self.BTN_ADD_ALL]
            removebuttons = [self.BTN_REMOVE, self.BTN_REMOVE_ALL]
            
            if not available: self.__DisableButtons(addbuttons)
            else: self.__EnableButtons(addbuttons)
            if not active: self.__DisableButtons(removebuttons)
            else: self.__EnableButtons(removebuttons)

    def __RebuildListbox(self, listbox, list):
        listbox.Clear()
        listbox.InsertItems(list, 0)

    def __RebuildCombobox(self, combobox, sets):
        setnames = sets.keys()
        setnames.sort()
        
        combobox.Clear()
        combobox.AppendItems(setnames)

    def __ShowUnsavedDataQuestion(self):
        r = wx.MessageBox(_("Unsaved data. Save now?"), _("Unsaved data"),
                          wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION)
        return r

    def __MarkCurrentSetModified(self, setname=None):
        if not setname: setname = self.comboboxSets.GetValue()
        
        # Quit if set is already modified
        if setname[-1] == "*": return

        # Delete old entry (setname without *)
        tuple = list(self.sets[setname])
        del self.sets[setname]
        
        # Create new entry with * at the end
        setname = setname + "*"
        self.sets[setname] = tuple
        
        self.__RebuildCombobox(self.comboboxSets, self.sets)
        self.comboboxSets.SetValue(setname)
            
        # Enable save button
        self.__EnableButtons([self.BTN_SAVE_SET])

    def __MarkCurrentSetUnmodified(self, setname=None, setsel=True): 
        if not setname: setname = self.comboboxSets.GetValue()
        
        # Quit if set is already unmodified
        if setname[-1] != "*": return

        # Delete old entry (setname with *)
        tuple = list(self.sets[setname])
        del self.sets[setname]

        # Create new entry without * at the end
        setname = setname[:-1]
        self.sets[setname] = tuple

        self.__RebuildCombobox(self.comboboxSets, self.sets)
        if setsel: self.comboboxSets.SetValue(setname)
        
        # Disable save button
        self.__DisableButtons([self.BTN_SAVE_SET])

    def __MakeAvailableList(self):
        sActive = set(self.__MakeActiveList())
        sAll = set(self.available.values())
        sAvailable = sAll - sActive
        l = list(sAvailable)
        l.sort()
        return l

    def __MakeActiveList(self):
        if not self.sets: return []
        
        # Get name of current set
        setname = self.comboboxSets.GetValue()
        
        l = []
        for x in self.sets[setname]:
            l.append(self.available[x])

        l.sort()
        return l

    def __EnableButtons(self, ids, enable=True):
        if not ids: return

        for id in ids:
            btn = self.FindWindowById(id)
            if btn:
                if enable: btn.Enable()
                else: btn.Disable()

    def __DisableButtons(self, ids):
        self.__EnableButtons(ids, False)
