import wx
import wx.aui
import dialogs.about
import tree
import images
import threading, os, sys, time
from taskbaricon import TaskBarIcon
from dialogs.set import SetDialog
from dialogs.servicesettings import ServiceSettingsDialog
import services.service as modService
from services.service import Service
from config import *
from updatethread import UpdateThread
import freemet.splash as fmsplash

_ = wx.GetTranslation

class Frame(wx.Frame):
    """
    The main window which contains a tree and a viewer (notebook) with tabs
    """
    
    def __init__(self):
        wx.Frame.__init__(self, None, -1, "FreeMet", size=(1000, 700),
                          style = wx.DEFAULT_FRAME_STYLE)
        self.Bind(wx.EVT_CLOSE, self.OnCloseWindow)
        
        # Dictionary mapping set names to a list of active services
        self.sets = {}
        
        # Dictionary mapping service key names to service objects
        self.services = {}
        
        # Reference to currently selected service object
        self.currentService = None
        
        # Name of current set
        self.currentSet = ""
        
        # List of services to load at program startup or selection of another set
        self.loadOnStartup = {}
        
        self.taskbaricon = TaskBarIcon(self)
        
        # Load config file "config/freemet.conf"
        self.__ReadConfigSets("config/freemet.conf")
        
        Service.configPath = "config/" + self.currentSet + "/"
        Service.downloadPath = "images/" + self.currentSet + "/"
        
        # Create splash screen before loading the set
        self.splash = fmsplash.MySplashScreen()
        splashWnd = self.splash.GetSplashWindow()
        
        self.__LoadSet(self.currentSet)
        
        self.splash.Close()
        
        # Mapping of service key names to their full names
        self.serviceNameMap = self.__GetServiceNameTuples()
        
        # Create a menu and a status bar
        self.CreateStatusBar(1, wx.ST_SIZEGRIP)
        self.CreateMenu()
        
        # Create a panel inside the frame and manage it with aui
        # mainpanel is a workaround for a mac bug
        self.mainpanel = wx.Panel(self)
        self.mgr = wx.aui.AuiManager(self.mainpanel)
        
        # Create viewer notebook for maps etc.
        self.notebook = wx.Notebook(self.mainpanel, -1)
        
        # Create tree for plugin selection
        leftPanel = self.CreateTree()
        
        # Assemble gui layout
        self.mgr.AddPane(self.notebook,
                         wx.aui.AuiPaneInfo().CenterPane().Name("Notebook"))
        self.mgr.AddPane(leftPanel,
                         wx.aui.AuiPaneInfo().
                         Left().Layer(2).BestSize((300, -1)).
                         MinSize((200, -1)).
                         Floatable(False).FloatingSize((300, 700)).
                         Caption("Maps Tree").
                         CloseButton(False))
        self.mgr.Update()
        
        # Start update threads for each service
        self.__StartUpdateThreads()
        
        # Create a timer which checks every second if service needs to be updated
        self.timer = wx.Timer(self)
        self.timer.Start(1000)
        self.Bind(wx.EVT_TIMER, self.OnTimer)
        
        self.Center(wx.BOTH)
        
    def CreateMenu(self):
        menuFile = wx.Menu()
        item1 = menuFile.Append(-1, _("&Exit\tCtrl-Q"),
                                _("Exit program"))
        self.Bind(wx.EVT_MENU, self.OnExit, item1)
        
        menuServices = wx.Menu()
        item = menuServices.Append(-1, _("&Update services"),
                                   _("Updates all weather services"))
        self.Bind(wx.EVT_MENU, self.OnMenuUpdateServices, item)
        item = menuServices.Append(-1, _("&Update current service\tF5"),
                                   _("Updates current services"))
        self.Bind(wx.EVT_MENU, self.OnMenuUpdateCurrentService, item)
        
        menuSettings = wx.Menu()
        item = menuSettings.Append(-1, _("Sets..."), _("Manage sets"))
        self.Bind(wx.EVT_MENU, self.OnMenuSettingsSets, item)
        item = menuSettings.Append(-1, _("Service settings..."),
                                   _("Update interval and startup behaviour"))
        self.Bind(wx.EVT_MENU, self.OnMenuServiceSettings, item)
        
        menuHelp = wx.Menu()
        item = menuHelp.Append(-1, _("&About..."),
                            _("Information about the program"))
        self.Bind(wx.EVT_MENU, self.OnMenuAbout, item)
        
        self.menubar = wx.MenuBar()
        self.menubar.Append(menuFile, _("&File"))
        self.menubar.Append(menuServices, _("S&ervices"))
        self.menubar.Append(menuSettings, _("&Settings"))
        self.menubar.Append(menuHelp, _("&Help"))
        self.SetMenuBar(self.menubar)
        
    def RebuildTree(self):
        self.tree.DeleteAllItems()
        self.serviceItems = {}
        
        active = self.sets[self.currentSet]
        
        root = self.tree.AddRoot(_("Overview"))
        
        # Create categories
        catItems = {}
        for cat, catName in modService.categories.iteritems():
            catItems[cat] = self.tree.AppendItem(root, catName)
        
        # Create service entries
        for key in active:
            service = self.services[key]
            assert service
            
            sName, fName = service.GetNameTuple()
            
            cat = service.GetCategory()
            if not cat or len(cat) == 0: parent = root
            else: parent = catItems[cat]
            
            text = fName
            if key in self.loadOnStartup[self.currentSet]:
                text += " (%s)" % time.strftime("%H:%M")
            
            self.serviceItems[key] = self.tree.AppendItem(parent, text)
            self.tree.SetPyData(self.serviceItems[key], (sName, fName))
            
        for item in catItems.itervalues():
            # Remove categories with no children
            if not self.tree.ItemHasChildren(item):
                self.tree.Delete(item)
                
            # Sort children of remaining category items
            else:
                self.tree.SortChildren(item)
                
        # Sort categories alphabetically
        self.tree.SortChildren(root)
        
        self.tree.ExpandAll()
        self.tree.SelectItem(root)
        
    def CreateTree(self):
        leftPanel = wx.Panel(self.mainpanel,
                             style=wx.TAB_TRAVERSAL | wx.CLIP_CHILDREN)
        self.tree = tree.MapsTree(leftPanel)
        w, h = self.GetClientSizeTuple()
        self.tree.SetDimensions(0, 0, 300, h)
        self.tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnSelectTreeItem)
        
        self.RebuildTree()
        
        return leftPanel
        
    def OnExit(self, evt):
        # Kill all child threads
        map(lambda t: t.Cancel(), self.updateThreads.itervalues())
        """
        for t in self.updateThreads.itervalues():
            pid = t.GetPid()
            #if config["verbose"]: print "Killing thread %i" % pid
            #os.system("kill " + str(pid))
        """
            
        # Destroy taskbar and the frame itself
        self.taskbaricon.Destroy()
        self.Destroy()
        
    def OnCloseWindow(self, evt):
        self.Hide()
        
    def OnTimer(self, evt):
        timerIntervalMs = evt.GetInterval()
        curTimeStr = time.strftime(config["timeFormat"])
        statusBar = self.GetStatusBar()
        
        while self.pendingSetStatusText:
            key = self.pendingSetStatusText.pop(0)
            service = self.services[key]
            sName, fName = service.GetNameTuple()
            statusBar.SetStatusText(_("Downloading") + " " + fName)
            
        while self.pendingReload:
            key = self.pendingReload.pop(0)
            service = self.services[key]
            sName, fName = service.GetNameTuple()
            
            statusBar.SetStatusText(_("Reloading") + " " + fName)
            service.RefreshPages(service == self.currentService)
            statusBar.SetStatusText(curTimeStr + ": " + fName + " " +
                                    _("completely updated"))
            
            # When service has been reloaded append update time to tree item
            item = self.serviceItems[key]
            text = self.tree.GetItemText(item)
            parts = text.split("(")
            if len(parts) > 1: text = ''.join(parts[0:-1])[:-1]
            self.tree.SetItemText(item, "%s (%s)" % (text, curTimeStr))
        
    def OnMenuChangeTbIcon(self, evt):
        self.taskbaricon.ChangeIcon()
    
    def OnMenuAbout(self, evt):
        d = dialogs.about.About()
        d.Show()
        
    def OnMenuUpdateServices(self, evt):
        for t in self.updateThreads.itervalues():
            t.Update()
        
    def OnMenuUpdateCurrentService(self, evt):
        sName, fName = self.currentService.GetNameTuple()
        self.updateThreads[sName].Update()
        
    def OnMenuSettingsSets(self, evt):        
        dlg = SetDialog(self, self.sets, self.currentSet, self.serviceNameMap)
        r = dlg.ShowModal()

        if r == wx.ID_OK:
            # Get parameters from dialog
            self.sets = dlg.GetSetList()
            self.currentSet = dlg.GetSetName()
            Service.configPath = "config/" + self.currentSet + "/"
            Service.downloadPath = "images/" + self.currentSet + "/"
            
            # Write new sets to config file
            self.__WriteConfigSets("config/freemet.conf")
            
            # Shutdown all update threads
            map(lambda t: t.Cancel(), self.updateThreads.values())
            
            # Reload all services
            self.__LoadSet(self.currentSet)
            
            # Restart update threads
            self.__StartUpdateThreads()
            
            # Rebuild selection tree and delete notebook pages
            self.RebuildTree()
            self.notebook.DeleteAllPages()

        dlg.Destroy()
        
    def OnMenuServiceSettings(self, evt):            
        dlg = ServiceSettingsDialog(self, self.services,#allServices,
                                    self.loadOnStartup[self.currentSet],
                                    self.currentSet)
        r = dlg.ShowModal()
        
        # Save new settings to config file
        self.__WriteConfigSets("config/freemet.conf")
        
        # Shutdown all update threads
        map(lambda t: t.Cancel(), self.updateThreads.values())
                
        # Restart update threads
        self.__StartUpdateThreads()
    
    def OnSelectTreeItem(self, evt):
        item = evt.GetItem()
        if item and not self.tree.ItemHasChildren(item):
            sName, fName = self.tree.GetPyData(item)

            # Remove pages instead of deleting them, so that the page objects
            # remain loaded in the memory.
            # The reason why page #0 is removed in each iteration is that
            # after deleting a page from a 3-page notebook for instance,
            # only page 0 and page 1 are left -- page 2 is gone.
            """
            for i in range(self.notebook.GetPageCount()):
                self.notebook.GetPage(0).Hide()
                self.notebook.RemovePage(0)
            """
            if self.currentService:
                self.currentService.RemovePages()
            
            self.currentService = self.services[sName]
            self.currentService.ShowNotebook(self.notebook)
            
    def GetConfigPath(self):
        return "config/" + self.currentSet + "/"
    
    def __GetServiceNameTuples(self):
        tuples = {}
        
        for key in self.__GetAvailableServices():
            deleteObjectAfterwards = False
            
            if key in self.services.keys():
                obj = self.services[key]
            else:
                mod = self.__ImportModule("freemet.services." + key)
                obj = mod.load()
                deleteObjectAfterwards = True
                
            assert obj
            sName, fName = obj.GetNameTuple()
            tuples[sName] = fName
            
            if deleteObjectAfterwards: del obj
            
        return tuples
        
    def __GetAvailableServices(self):
        def GetModules(folder, exclude=None):
            assert folder
            assert len(folder) > 0
            
            if folder[-1] != "/": folder += "/"
            
            # Read py and pyc files from the directory
            from glob import glob
            files = glob(folder + "*.py")
            files += glob(folder + "*.pyc")
            del glob
            
            # Quit function if there are no files or path was wrong
            if not files: return [] 
            
            # Remove path prefix
            filenames = map(lambda s: s.split("/")[-1], files)
            
            # Remove file extension and generate set (no duplicate entries)
            modules = set(map(lambda s: s.split(".")[-2], filenames))
            
            # Remove unwanted modules (e.g. __init__)
            if exclude:
                for excl in exclude:
                    if excl in modules: modules.remove(excl)
            
            return list(modules)
            
        # Search services folder for plugins
        modules = GetModules("freemet/services",
                             exclude=["__init__", "overview", "service"])
        
        if not modules:
            if config["verbose"]:
                print "No services found to import"

        return modules
    
    def __ImportServicesModules(self, modules):
        if not modules: return
        
        # Import service modules
        rawServices = []
        for modstring in modules:
            mod = self.__ImportModule(modstring)
            rawServices.append(mod.load())
            
        # Load and initialize services
        self.services = {}
        numDownloads = 0
        for service in rawServices:
            sName, fName = service.GetNameTuple()
            self.services[sName] = service
            
    def __ReadConfigSets(self, filename):
        assert filename
        self.__config = wx.FileConfig(localFilename=filename,
                                      style=wx.CONFIG_USE_LOCAL_FILE
                                      | wx.CONFIG_USE_RELATIVE_PATH)
        assert self.__config
        
        self.sets = {}
        
        def Deserialize(s): return s.split(",")
        
        # Read set names
        setnames = Deserialize(self.__config.Read("sets/names"))
        self.currentSet = self.__config.Read("sets/currentSet")
        
        for setname in setnames:
            cfgSet = wx.FileConfig(localFilename="config/" + setname + "/set.cfg",
                                   style=wx.CONFIG_USE_LOCAL_FILE
                                   | wx.CONFIG_USE_RELATIVE_PATH)
            assert cfgSet
            
            active = Deserialize(cfgSet.Read("active"))
            loadOnStartup = Deserialize(cfgSet.Read("loadOnStartup"))
            
            if active == [""]: active = []
            self.sets[setname] = active
            
            if loadOnStartup == [""]: loadOnStartup = []
            self.loadOnStartup[setname] = loadOnStartup
            
            del cfgSet
            
    def __WriteConfigSets(self, filename):
        assert filename
        assert self.__config
        
        if not self.sets: return
        
        def Serialize(l):
            s = ""
            for x in l: s += x + ","
            return s[:-1]

        setnames = []
        
        for setname, active in self.sets.iteritems():
            if setname not in setnames: setnames.append(setname)
            
            cfgSet = wx.FileConfig(localFilename="config/" + setname + "/set.cfg",
                                   style=wx.CONFIG_USE_LOCAL_FILE
                                   | wx.CONFIG_USE_RELATIVE_PATH)
            assert cfgSet
            
            serActive = Serialize(active)
            serLoadOnStartup = Serialize(self.loadOnStartup[setname])
            cfgSet.Write("active", serActive)
            cfgSet.Write("loadOnStartup", serLoadOnStartup)
            cfgSet.Flush()
            del cfgSet
            
        setnames.sort()
        self.__config.Write("sets/names", Serialize(setnames))
        self.__config.Write("sets/currentSet", self.currentSet)
        
        self.__config.Flush()
        
    def __ImportModule(self, name):
        mod = __import__(name)
        components = name.split('.')
        for comp in components[1:]:
            mod = getattr(mod, comp)
        
        return mod
        
    def __LoadSet(self, setname):
        # Load services in the current set
        setModules = list(self.sets[setname])
        if setModules:
            setModules = map(lambda s: "freemet.services." + s, setModules)
            self.__ImportServicesModules(setModules)
        
        # Count downloads (for progress bar)
        numDownloads = 0
        for key, service in self.services.iteritems():
            if key in self.loadOnStartup[setname]:
                numDownloads += service.GetNumDownloads()
            
        # Create progress dialog
        """
        progressDlg = wx.ProgressDialog(_("Loading ..."), "",
                                        maximum=numDownloads, parent=self,
                                        style=wx.PD_CAN_ABORT
                                        | wx.PD_AUTO_HIDE
                                        | wx.PD_APP_MODAL
                                        | wx.PD_ELAPSED_TIME
                                        | wx.PD_ESTIMATED_TIME
                                        | wx.PD_REMAINING_TIME)
        """
        progressDlg = None

        # Load services
        count = 0
        skipFurtherDownloads = False
        for key, obj in self.services.iteritems():
            # Skip download of this service if it should not be loaded right now
            download = key in self.loadOnStartup[self.currentSet] and True or False
            
            if skipFurtherDownloads: download = False
            
            print "DrawText", key
            self.splash.DrawText(_("Loading") + " " + obj.GetNameTuple()[1])
            
            numLoaded, cancelBtnPressed = obj.Load(download, progressDlg, count)
            if cancelBtnPressed: skipFurtherDownloads = True
            
            count = count + numLoaded
            
        # Finally, destroy progress dialog
        #progressDlg.Destroy()
        
    def __StartUpdateThreads(self):
        self.pendingReload = []
        self.pendingSetStatusText = []
        self.updateThreads = {}
        for key, obj in self.services.iteritems():
            self.updateThreads[key] = UpdateThread(self, obj)
            self.updateThreads[key].start()