"""
The service superclass
"""

import urllib2, os, socket
import wx
from thread import *
from freemet.config import config

# Do not abort program if Image module is not available -- saving GIF files is
# not a critical function.
try: import Image
except: pass

_ = wx.GetTranslation

# A 60 seconds timeout seems to be fine even for the slow Wellington Views website
socket.setdefaulttimeout(60)

categories = {"forecast": _("Forecast"),
              "maps": _("Maps"),
              "satellite": _("Satellite"),
              "webcams": _("Webcams"),
              "archive": _("Archive")}

class Service:
    """
    The service superclass
    
    A service is the loading, processing and providing of a map, a series of
    maps or some other weather related values.
    The simplest example is just to load and show an image file. A more
    sophisticated example would be parsing text and present only the most
    important informations to the user by building a diagram.
    This class is intended to be subclassed by each new source of information.
    """
    
    notebook = None
    configPath = ""
    downloadPath = ""
    
    def __init__(self, shortName, fullName, numDownloads, updateInterval,
                 pages, category):
        """
        Constructs an service object
        
        The constructor expects to get the service's short name (for internal
        purposes like dictionaries), the full name (shown in the service
        selection tree), the number of downloads (for the progress bar), the
        update interval (0 = never automatically), a list of page names, and
        the category.
        
        Subclassing a service is quite easy: Just create a new class, call
        Service.__init__ from your __init__ method. There are two more methods
        you have to implement: CreatePage() and Load().
        CreatePage: must return a wx.Window object which will be put into the
        notebook page with the number given by the parameter index.
        Load: you can call Service.Download() and pass a list of urls
        """
        
        self.__panels = []
        self.__mutex = allocate_lock()
         
        self.__shortName        = shortName
        self.__fullName         = fullName
        self.__numDownloads     = numDownloads
        self.__updateInterval   = updateInterval
        self.__pages            = pages
        self.__category         = category
        
        # Necessary for popup menu
        self.__files            = []
        self.__images           = {}
        
        self.__LoadConfig()

    def SetNumDownloads(self, numDownloads):
        self.__numDownloads = numDownloads
        
    def SetPages(self, pages):
        self.__pages = pages
    
    def Download(self, download, downloads, progressDlg=None, progressCount=0):
        """
        Downloads data as given in the downloads list as [(url, file), ...]
        """
        
        if not downloads: return 0, False
        
        self.__files = downloads
        
        if not download: return 0, False
        
        i = 0
        for url, file in downloads:
            if progressDlg:
                # Word wrapping
                url2 = url
                wrapwidth = 36
                for l in range(wrapwidth, 256, wrapwidth):
                    if len(url2) > l: url2 = url2[:l] + "\n" + url[l:]
                
                update, skip = progressDlg.Update(progressCount + i, url2)
                i += 1
                
                # User clicked Cancel button
                if not update:
                    return i, True
        
            if config["verbose"]:
                print self.__shortName + ": downloading " + url
            
            self.url2file(url, file)
            
        # Return number of downloaded files
        return i, False
    
    def ShowNotebook(self, nb):
        """
        Builds all the pages and adds them to the notebook
        
        Subclasses must provide a function CreatePage(index), which returns
        a wx.Window object to be placed into the notebook page.
        """
        
        self.notebook = nb
        Service.notebook = nb
        self.notebook.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGING,
                           self.OnNotebookPageChanging)
        
        if self.__panels:
            for page in self.__pages:
                idx = self.__pages.index(page)
                
                nb.AddPage(self.__panels[idx], page)#, imageId=idx)
                self.notebook.GetPage(idx).Show()
            
            self.__panels[0].Show()
            
            return
        
        #il = wx.ImageList(16, 16)
        #for png in ["overview", "code", "demo"]:
        #    bmp = images.catalog[png].GetBitmap()
        #    il.Add(bmp)
        #self.notebook.AssignImageList(il)
        
        for page in self.__pages:
            idx = self.__pages.index(page)
            p = self.CreatePage(idx)
            assert p
            self.__panels.append(p)
            nb.AddPage(self.__panels[idx], page)#, imageId=idx)
            
    def RemovePages(self):
        for i in range(self.notebook.GetPageCount()):
            self.__panels[i].Hide()
            self.notebook.GetPage(0).Hide()
            self.notebook.RemovePage(0)
    
    def GetNameTuple(self):
        return self.__shortName, self.__fullName
    
    def GetNumDownloads(self):
        return self.__numDownloads
    
    def GetUpdateInterval(self):
        return self.__updateInterval
    
    def SetUpdateInterval(self, i):
        self.__updateInterval = i
    
    def GetCategory(self):
        return self.__category
    
    def url2file(self, url, file):
        """
        Provides an easy way to download an url to a local file
        
        Subclasses usually don't need this because Service.Download() only
        needs a list of urls to download everything.
        """
        
        if config["disable_download"]:
            if config["verbose"]:
                print "skipping download of " + url + " due to -ddownload flag"  
            return True
        
        assert file != None
        assert file != ""
        assert url != None
        assert url != ""
        
        userAgent = 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)'
        headers = {"User-Agent": userAgent}
        
        try:
            # Download via http
            req = urllib2.Request(url, headers=headers)
            response = urllib2.urlopen(req)
            s = response.read()
            response.close()

        except urllib2.URLError, e:
            print "Error downloading " + url
            print "Error message: " + e.message
            return False
        
        # Lock mutex for writing downloaded image to a file
        self.__mutex.acquire()
        
        # Build path = set dir + service dir
        downloadPath = self.downloadPath + self.__shortName + "/"
        
        # Create download directory (try/except in case directory exists already)
        try:
            os.makedirs(downloadPath)
        except OSError:
            pass
        
        try:
            # Write image to freemet folder
            fout = open(downloadPath + file, "wb")
            fout.write(s)
            fout.close()
        except IOError, (errno, errstr):
            print "Error writing image " + downloadPath + file + "(" + errstr + ")"
            self.__mutex.release()
            return False
        
        # Release image file mutex
        self.__mutex.release()
        
        return True
    
    def AddImageNotebookPage(self, parent, file):
        """
        Adds a static image to the parent window (multi-threading safe)
        """
        
        assert parent != None
        assert file != None
        assert file != ""
        
        path = self.downloadPath + self.__shortName + "/" + file
        
        self.__mutex.acquire()
        img = wx.Image(path)
        self.__images[file] = img
        self.__mutex.release()
        
        try:
            bmp = wx.StaticBitmap(parent, -1, img.ConvertToBitmap())
        except:
            bmp = wx.StaticBitmap(parent, -1)
        
        bmp.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        return bmp
    
    def AddAnimatedImageNotebookPage(self, parent, file):
        """
        Adds an animated image to the parent window (multi-threading safe)
        """
        
        assert parent != None
        assert file != None
        assert file != ""
        
        path = self.downloadPath + self.__shortName + "/" + file
        
        # Protect loading of image with a mutex because the update thread might
        # download a new image at the same moment.
        self.__mutex.acquire()
        try:
            gif = wx.animate.GIFAnimationCtrl(parent, -1, path)
        except:
            gif = None
        
        ani = gif.GetAnimation()
        self.__images[file] = ani
        self.__mutex.release()
        
        gif.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        return gif
    
    def OnPopupMenuSaveFile(self, file, img):
        if img == None:
            wx.MessageBox(_("No image selected to save to file"), _("Error"),
                          wx.ICON_ERROR)
            return
        
        defFile = file
        extension = defFile.split(".")[-1].lower()
        if extension == "bmp":
            imgtype = wx.BITMAP_TYPE_BMP
        elif extension == "png":
            imgtype = wx.BITMAP_TYPE_PNG
        elif extension == "jpg" or extension == "jpeg":
            imgtype = wx.BITMAP_TYPE_JPEG
        elif extension == "tif":
            imgtype = wx.BITMAP_TYPE_TIF
        elif extension == "gif":
            imgtype = wx.BITMAP_TYPE_GIF   
            # Image module not found and therefore not imported
            try: Image
            except:
                wx.MessageBox(_("Image module not found,\ncannot save GIF files."),
                              _("Error"),
                              wx.ICON_ERROR)
            return           
            
        wildcard = "*." + extension
        if not defFile or not extension or not wildcard:
            defFile = ""
            wildcard = ""
            
        # Get user's home directory
        defDir = os.path.expanduser("~")
        
        # Create save file dialog
        dlg = wx.FileDialog(self.notebook, _("Save file as ..."),
                            defaultDir=defDir,
                            defaultFile=defFile, wildcard=wildcard,
                            style=wx.SAVE | wx.OVERWRITE_PROMPT)
        
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            
            # wx cannot save gif files, so use Python's Image module
            # instead
            if imgtype == wx.BITMAP_TYPE_GIF:
                img = Image.open(self.downloadPath + self.__shortName +
                                 "/" + defFile)
                img.save(path)
                del img
            else:
                img.SaveFile(path, imgtype)
            
        dlg.Destroy()
    
    def OnContextMenu(self, evt):
        curPageIdx = self.notebook.GetSelection()
        page = self.notebook.GetCurrentPage()
        mousePos = page.ScreenToClient(evt.GetPosition())
        
        # Create standard popup menu
        menu = wx.Menu()
        menu.Append(0, _("Save image..."))
        menu.Append(1, _("Copy URL"))
        
        def OnPopupMenu(evtMenu):            
            if evtMenu.GetId() == 0:
                file = self.__files[curPageIdx][1]
                img = self.__images[file]                
                self.OnPopupMenuSaveFile(file, img)
            
            elif evtMenu.GetId() == 1:
                data = wx.TextDataObject()
                data.SetText(self.__files[curPageIdx][0])
                if wx.TheClipboard.Open():
                    wx.TheClipboard.SetData(data)
                    wx.TheClipboard.Close()
        
        # Bind click on a menu item to OnPopupMenu 
        menu.Bind(wx.EVT_MENU, OnPopupMenu)

        # Show popup menu with notebook as parent window
        self.notebook.PopupMenu(menu)
        
        # Destroy the menu
        menu.Destroy()
    
    def RefreshPages(self, show):
        """
        Deletes and rebuilds all notebook pages
        """
        
        #if self.notebook:
        # Delete everything
        self.__panels = []
    
        # Show notebook pages with new images
        if show:
            self.notebook.DeleteAllPages()
            self.ShowNotebook(self.notebook)
            
    def OnNotebookPageChanging(self, evt):
        oldSel = evt.GetOldSelection()
        if oldSel != -1:
            self.DeactivateNotebookPage(oldSel)
            
        self.ActivateNotebookPage(evt.GetSelection())
        
    def ActivateNotebookPage(self, sel):
        """
        Called when a page is selected---most services with static data only
        don't need this
        """
        pass
    
    def DeactivateNotebookPage(self, sel):
        """
        Called when a page is deselected---most services with static data only
        don't need this
        """
        pass
    
    def ReadConfigString(self, key):
        return self.__config.Read(key)
    
    def ReadConfigBool(self, key):
        return self.__config.ReadBool(key)
    
    def ReadConfigInt(self, key):
        return self.__config.ReadInt(key)
    
    def ReadConfigFloat(self, key):
        return self.__config.ReadFloat(key)
    
    def WriteConfigString(self, key, value):
        self.__config.Write(key, value)
    
    def WriteConfigBool(self, key, value):
        self.__config.WriteBool(key, value)
    
    def WriteConfigInt(self, key, value):
        self.__config.WriteInt(key, value)
    
    def WriteConfigFloat(self, key, value):
        self.__config.WriteFloat(key, value)
        
    def FlushConfigFile(self):
        self.__config.Flush()
        
    def GetConfigGroup(self, index):
        return self.__config.GetNextGroup(index)
    
    def DeleteConfigGroup(self, path):
        return self.__config.DeleteGroup(path)
        
    def CreateAboutPage(self, description, author="", license=""):
        """
        Provides an easy way to build a page with information about the
        service---a quite unfancy one, though. 
        """
        
        assert description != None
        assert author != None
        assert license != None
        
        p = wx.Panel(Service.notebook, -1)
        str = description + "\n\n" + _("Author: ") + author + \
            "\n" + _("License: ") + license
        wx.StaticText(p, -1, str, (10, 10))
        
        return p
    
    def SaveConfig(self):
        assert self.__config
        
        self.__config.WriteInt("updateInterval", self.__updateInterval)
        self.__config.Flush()
    
    def __LoadConfig(self):
        assert self.configPath != ""
        filename = self.configPath + self.__shortName + ".conf"
        if config["verbose"]:
            print "Parsing config file " + filename
            
        self.__config = wx.FileConfig(localFilename=filename,
                                      style=wx.CONFIG_USE_LOCAL_FILE
                                      | wx.CONFIG_USE_RELATIVE_PATH)
        assert self.__config
        
        self.__updateInterval = int(self.__config.ReadInt("updateInterval"))