import wx
import wx.grid
import hashlib
from freemet.services.service import Service

_ = wx.GetTranslation

class SettingsRow():
    def __init__(self, parent, sizer, webcam):
        assert parent
        assert sizer
        assert webcam
        
        self.parent = parent
        self.webcam = webcam
        
        # Create items
        self.textctrlName   = wx.TextCtrl(parent, value=webcam.name,
                                          size=(120,-1))
        self.textctrlUrl    = wx.TextCtrl(parent, value=webcam.url,
                                          size=(450,-1))
        self.btnUp          = wx.Button(parent, label="^", size=(30,-1))
        self.btnDown        = wx.Button(parent, label="v", size=(30,-1))
        self.btnDelete      = wx.Button(parent, label="X", size=(30,-1))
        
        if not self.webcam.prev:
            self.btnUp.Disable()
        if not self.webcam.next:
            self.btnDown.Disable()
        
        # Set bindings
        for btn in self.btnUp, self.btnDown, self.btnDelete:
            btn.Bind(wx.EVT_BUTTON, self.OnButton)

        self.textctrlName.Bind(wx.EVT_TEXT, self.OnCtrlUpdate)
        self.textctrlUrl.Bind(wx.EVT_TEXT, self.OnCtrlUpdate)
            
        # Add items to grid sizer
        l = [self.textctrlName, self.textctrlUrl,
             self.btnUp, self.btnDown, self.btnDelete]
        map(lambda x: sizer.Add(x), l)
        
    def OnButton(self, evt):
        id = evt.GetId()
        
        if id == self.btnUp.GetId():
            self.parent.MoveRowUp(self.webcam.id)
        
        elif id == self.btnDown.GetId():
            self.parent.MoveRowDown(self.webcam.id)
        
        elif id == self.btnDelete.GetId():
            self.parent.DeleteRow(self.webcam.id)
            
    def OnCtrlUpdate(self, evt):
        id = evt.GetId()
        
        if self.textctrlName.GetId() == id:
            self.webcam.name = self.textctrlName.GetValue()
        elif self.textctrlUrl.GetId() == id:
            self.webcam.url = self.textctrlUrl.GetValue()
       
class Webcam():
    def __init__(self, id, name, url, position, prev=None, next=None):
        self.id         = id
        self.name       = name
        self.url        = url
        self.position   = position
        self.prev       = prev
        self.next       = next
            
class WebcamLinkedList():
    """
    A double linked list
    """
    
    def __init__(self, r=None):
        if not r or r.head == None:
            self.head = None
            self.tail = None
            return
        
        webcam = r.head
        prev = None
        while webcam:
            new = Webcam(webcam.id, webcam.name, webcam.url, webcam.position)
            
            if prev: prev.next = new
            else: self.head = new
            
            new.prev = prev
            prev = new
            
            webcam = webcam.next
        
        self.tail = prev
        
        assert self.head.prev == None
        assert self.tail.next == None
        
    def Append(self, webcam):
        if not self.head:
            self.head = webcam
            self.tail = webcam
            
        assert self.tail
        
        prev = self.tail
        prev.next = webcam
        
        self.tail = webcam
        self.tail.prev = prev
        self.tail.next = None
        
    def Delete(self, webcam):
        if not self.head:
            assert not self.tail
            return
        
        # Change pointers
        if webcam.prev:
            webcam.prev.next = webcam.next
        if webcam.next:
            webcam.next.prev = webcam.prev
            
        if webcam == self.head:
            self.head = webcam.next
        if webcam == self.tail:
            self.tail = webcam.prev
        
    def Clear(self):
        x = self.head
        while x:
            next = x.next
            del x
            x = next
            
        self.head = self.tail = None
        
    def Count(self):
        x = self.head
        num = 0
        while x:
            num += 1
            x = x.next
            
        return num
    
    def Swap(self, a, b):
        # Make sure that a is above b
        if a.position > b.position:
            t = a
            a = b
            b = t
        
        # Swap position variable
        tmp = a.position
        a.position = b.position
        b.position = tmp
        
        # Change links of swapped items
        id1Prev = a.prev
        id2Next = b.next
        a.prev = b
        b.next = a
        a.next = id2Next
        b.prev = id1Prev
        
        # Change links of items above and below
        if id1Prev: id1Prev.next = b
        if id2Next: id2Next.prev = a
        
        # Change head if appropriate
        # not id1Prev means that there was no row above a (now b)
        if not id1Prev: self.head = b

class SettingsPanel(wx.Panel):
    def __init__(self, service, webcams):
        wx.Panel.__init__(self, Service.notebook, -1)
        
        # Reference to caller
        self.service = service
        
        self.rows = []

        # Dictionary: id -> (name, url, position)
        self.webcams = WebcamLinkedList(webcams)
        
        # Set idCounter to greatest used id + 1
        maxId = -1
        webcam = self.webcams.head
        while webcam:
            if webcam.id > maxId: maxId = webcam.id
            webcam = webcam.next
        self.idCounter = maxId + 1
            
        self.savedWebcams = WebcamLinkedList(self.webcams)

        # Create sizers
        self.mainsizer = wx.BoxSizer(wx.VERTICAL)
        self.gridsizer = wx.FlexGridSizer(0, 5, 2, 2)
        buttonsizer = wx.BoxSizer(wx.HORIZONTAL)
        
        # Add a settings row for each webcam
        self.mainsizer.Add(self.gridsizer)
        self.__RebuildRows()
        
        # Put a button underneath the rows for adding a new webcam
        btn = wx.Button(self, wx.ID_ADD)
        self.Bind(wx.EVT_BUTTON, self.OnButton, btn)
        self.mainsizer.Add(btn)
        
        # Create undo and apply buttons
        for btnId in [wx.ID_UNDO, wx.ID_APPLY]:
            btn = wx.Button(self, btnId)
            #btn.Disable()
            self.Bind(wx.EVT_BUTTON, self.OnButton, btn)   
            buttonsizer.Add(btn, 0, wx.ALIGN_BOTTOM | wx.ALL, 5)
        
        # Put everything together
        self.mainsizer.Add(buttonsizer, 1, wx.ALIGN_BOTTOM | wx.ALIGN_RIGHT)
        self.SetSizer(self.mainsizer)
        
    def OnButton(self, evt):
        id = evt.GetId()
        
        if id == wx.ID_APPLY:
            self.savedWebcams = WebcamLinkedList(self.webcams)
            self.service.OnPropertiesApply(self.savedWebcams)
        
        elif id == wx.ID_UNDO:
            self.webcams = WebcamLinkedList(self.savedWebcams)
            self.__RebuildRows()
        
        elif id == wx.ID_ADD:
            id = self.idCounter
            self.idCounter += 1
            
            print self.webcams.tail
            print self.webcams.tail.position
            
            position = 0
            if self.webcams.tail:
                position = self.webcams.tail.position + 1

            webcam = Webcam(id, _("New"), "", position)
            self.webcams.Append(webcam)
            
            self.__RebuildRows()
        
    def MoveRowUp(self, id):
        webcam = self.webcams.head
        while webcam:
            if webcam.id == id: break
            webcam = webcam.next
        
        assert webcam
        if not webcam.prev: return
        
        self.webcams.Swap(webcam.prev, webcam)
        self.__RebuildRows()
        
    def MoveRowDown(self, id):
        webcam = self.webcams.head
        while webcam:
            if webcam.id == id: break
            webcam = webcam.next
        
        assert webcam
        if not webcam.next: return
        
        self.webcams.Swap(webcam, webcam.next)
        self.__RebuildRows()
        
    def DeleteRow(self, id):
        webcam = self.webcams.head
        while webcam:
            if webcam.id == id: break
            webcam = webcam.next
            
        assert webcam
        
        # Change position variable of all webcams below
        # We need the variable webcam later thus use webcam2 for position
        # changing.
        webcam2 = webcam.next
        while webcam2:
            webcam2.position -= 1
            webcam2 = webcam2.next
            
        self.webcams.Delete(webcam)
        
        # Rebuild settings gui
        self.__RebuildRows()
        
    def __RebuildRows(self):
        self.gridsizer.Clear(True)
        del self.rows[:]
        
        # Add labels for each column
        self.gridsizer.Add(wx.StaticText(self, -1, _("Name:")))
        self.gridsizer.Add(wx.StaticText(self, -1, _("URL:")))
        self.gridsizer.AddSizer((1,1)) # Column for button ^
        self.gridsizer.AddSizer((1,1)) # Column for button v        
        self.gridsizer.AddSizer((1,1)) # Column for button X
        
        webcam = self.webcams.head
        while webcam:
            row = SettingsRow(self, self.gridsizer, webcam)
            self.rows.append(row)
            webcam = webcam.next
            
        self.gridsizer.Layout()
        self.mainsizer.Layout()
        
class Webcams(Service):
    def __init__(self):
        Service.__init__(self, "webcams", _("Webcams"), 0, 1, [],
                         "webcams")
        
        self.webcams = WebcamLinkedList()
        
        self.__ReadConfigFile()
        
        # Get downloads number by counting all urls (with ;)
        webcam = self.webcams.head
        numDownloads = 0
        while webcam:
            numDownloads += len(webcam.url.split(";"))
            webcam = webcam.next
        
        self.SetNumDownloads(numDownloads)
        
        # Get page titles (in correct order)
        pages = []
        webcam = self.webcams.head
        while webcam:
            pages.append(webcam.name)
            webcam = webcam.next
            
        pages += [_("Settings"), _("About")]
        self.SetPages(pages)
        
    def Load(self, download, progressDlg=None, progressCount=0):
        downloads = []
        webcam = self.webcams.head
        while webcam:
            # Split urls if necessary
            for spliturl in webcam.url.split(";"):
                file = self.__GenerateFileName(spliturl)
                downloads.append((spliturl, file))
            
            webcam = webcam.next
        
        return self.Download(download, downloads, progressDlg, progressCount)
    
    def CreatePage(self, index):
        # Webcam pages go from 0 to |webcams|-1
        idxSettingsPage = self.webcams.Count()
        idxAboutPage = idxSettingsPage + 1
        
        if index == idxSettingsPage:
            p = SettingsPanel(self, self.webcams)
            
        elif index == idxAboutPage:
            descr = _("This service downloads webcam images from several\n"\
                      "websites (see settings).")
            p = self.CreateAboutPage(descr, "(c) 2008 Tobias Wulff", "GPL 2")
            
        else:
            p = wx.ScrolledWindow(Service.notebook, -1)
            p.SetScrollRate(10, 10)
            sizer = wx.BoxSizer(wx.HORIZONTAL)
            
            webcam = self.webcams.head
            while webcam:
                if webcam.position == index: break
                webcam = webcam.next
                
            assert webcam
                
            for url in webcam.url.split(";"):
                file = self.__GenerateFileName(url)
                sizer.Add(self.AddImageNotebookPage(p, file))
                
            p.SetSizer(sizer)
            
        return p
    
    def OnPropertiesApply(self, webcams):
        self.webcams = WebcamLinkedList(webcams)
        
        # Write new settings to config file
        self.__WriteConfigFile()
        
        # Get page titles (in correct order)
        pages = []
        webcam = self.webcams.head
        while webcam:
            pages.append(webcam.name)
            webcam = webcam.next
        pages += [_("Settings"), _("About")]
        self.SetPages(pages)
        
        # Reload all images
        self.Load(True)
        self.RefreshPages(True)
        
    def OnContextMenu(self, evt):
        curPageIdx = self.notebook.GetSelection()
        page = self.notebook.GetCurrentPage()
        mousePos = page.CalcUnscrolledPosition(page.ScreenToClient(evt.GetPosition()))
        scrollPos = page.CalcUnscrolledPosition(0, 0)
        
        # Get filenames of images on current page
        filenames = []
        urls = []
        webcam = self.webcams.head
        while webcam:
            if webcam.position == curPageIdx:
                for url in webcam.url.split(";"):
                    filenames.append(self.__GenerateFileName(url))
                    urls.append(url)
            
            webcam = webcam.next

        # Get image object under mouse cursor
        img = None
        file = ""
        url = ""
        children = page.GetChildren()
        for i, bmp in enumerate(children):
            winPos = bmp.GetPosition()
            winSize = bmp.GetSize()
            if mousePos.x >= winPos.x and \
               mousePos.x < winPos.x + winSize.width:
                img = bmp.GetBitmap().ConvertToImage()
                file = filenames[i]
                url = urls[i]
                break
        
        # Create standard popup menu
        menu = wx.Menu()
        menu.Append(0, _("Save image..."))
        menu.Append(1, _("Copy URL"))
        
        def OnPopupMenu(evtMenu):
            if evtMenu.GetId() == 0:
                self.OnPopupMenuSaveFile(file, img)
            
            elif evtMenu.GetId() == 1:
                data = wx.TextDataObject()
                data.SetText(url)
                if wx.TheClipboard.Open():
                    wx.TheClipboard.SetData(data)
                    wx.TheClipboard.Close()
        
        # Bind click on a menu item to OnPopupMenu 
        menu.Bind(wx.EVT_MENU, OnPopupMenu)

        # Show popup menu with notebook as parent window
        self.notebook.PopupMenu(menu)
        
        # Destroy the menu
        menu.Destroy()
    
    def __GenerateFileName(self, url):
        m = hashlib.md5()
        m.update(url)
        return "webcam%s.jpg" % str(m.hexdigest())
    
    def __ReadConfigFile(self):
        webcamList = []
        self.webcams.Clear()
        
        cont, value, idx = self.GetConfigGroup(0)
        while cont:
            # Get number from string "id[0-9]"
            id = int(value[2:])
            path = "id%i/" % id
            
            # Read stored values
            name        = self.ReadConfigString(path + "name")
            url         = self.ReadConfigString(path + "url")
            position    = self.ReadConfigInt(path + "position")            
            
            # Create a new webcam object to store information like url and name
            webcam = Webcam(id, name, url, position, None, None)
            webcamList.append(webcam)
            
            # Get next group (i.e. next webcam)
            cont, value, idx = self.GetConfigGroup(idx)
            
        # Return if there are no webcams
        if not webcamList: return
        
        # Sort webcams by position and set prev/next pointers
        webcamList.sort(lambda a,b: a.position - b.position)
        
        map(lambda x: self.webcams.Append(x), webcamList)
    
    def __WriteConfigFile(self):
        # Delete all webcam settings
        cont, value, idx = self.GetConfigGroup(0)
        while cont:
            assert value
            self.DeleteConfigGroup(value)
            cont, value, idx = self.GetConfigGroup(idx)
        
        # Write settings for each webcam to config file
        webcam = self.webcams.head
        while webcam:
            path = "id%i/" % webcam.id
            self.WriteConfigString(path + "name", webcam.name)
            self.WriteConfigString(path + "url", webcam.url)
            self.WriteConfigInt(path + "position", webcam.position)
            
            webcam = webcam.next
        
        self.FlushConfigFile()
    
def load(): return Webcams()